#define TARGET_WIN32  

#include <AppKit/NSImageRep.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSWindow.h>
#include <AppKit/NSAffineTransform.h>
#include <Foundation/NSException.h>
#include <limits.h>
#include "vssym32.h"

#include "WIN32Theme_win32.h"
#include "WIN32VSImageRep.h"

RECT WIN32RectFromFrame(NSRect r);
NSRect GSFrameFromWIN32Rect(RECT rect);

/**
  * Import the internal WIN32 function from the windows backend.
  * This is probably unsafe, and we would want a more public
  * API in the future.
  */
@interface NSObject (BackendHacking)
// WIN32GState  
- (HDC) getHDC;
- (void) releaseHDC:(HDC)hdc;
- (int) windowNumber;

// GSGraphicsContext
- (id) currentGState;
@end


@implementation WIN32Theme (win32)

+ (void) initialize
{
  NSLog(@"Inserting WIN32VSImageRep class for themed images.");
  [NSImageRep registerImageRepClass:[WIN32VSImageRep class]];
}

- (id) initWithBundle:(NSBundle*)bundle
{
  self = [super initWithBundle:bundle];
  if (!self)
    return nil;
  NSLog(@"Initializing theme.");

  OSVERSIONINFO osvi;
  ZeroMemory(&osvi, sizeof(OSVERSIONINFO));
  osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

  GetVersionEx(&osvi);

  if ((osvi.dwMajorVersion > 5) || ((osvi.dwMajorVersion==5) && (osvi.dwMinorVersion>=1)))
    platformSupportsVS = YES;
  else
    platformSupportsVS = NO;
    
  userEnabledVS = YES;
  return self;
}

- (BOOL) isThemeActive
{
  if (IsThemeActive())
    return YES;
  else
    return NO;
}

- (HTHEME) themeWithClassName:(NSString*)className
{
  if ([self useVisualStyles])
  {
    HWND hWnd;
    HTHEME hTheme;
    wchar_t* classNameChars;

    hWnd = (HWND)[[[NSView focusView] window] windowNumber];
    classNameChars = objc_calloc([className length]+1, sizeof(wchar_t));
    [className getCharacters:classNameChars];
    
    hTheme = OpenThemeData(hWnd, classNameChars);
    
    objc_free(classNameChars);
    return hTheme;
  }
  return NULL;
}

- (void) releaseTheme:(HTHEME)hTheme
{
  CloseThemeData(hTheme);
}

- (NSString*) themeClassName:(HTHEME)hTheme
{
  WCHAR *nameBuffer;
  NSString *name;

  nameBuffer = objc_malloc(sizeof(WCHAR) * 2048);
  if (FAILED(GetThemeString(hTheme, 0, 0, TMT_NAME, nameBuffer, 2048)))
  {
    objc_free(nameBuffer);
    [NSException raise:WIN32ThemeException
               format:@"Error retrieving theme class name in -[WIN32Theme themeClassName:]"];
    return nil;
  }
  name = [[NSString alloc] initWithCharactersNoCopy:nameBuffer 
                                             length:wcslen(nameBuffer)
                                          freeWhenDone:YES];
  if (name == nil)
  {
    objc_free(nameBuffer);
    [NSException raise:WIN32ThemeException
                format:@"Error creating string for -[WIN32Theme themeClassName:]"];
  } 
  return AUTORELEASE(name);
}

- (BOOL) isTheme:(HTHEME)hTheme partDefined:(int)part
{
  if (IsThemePartDefined(hTheme, part, 0))
    return YES;
  else 
    return NO;
}  

- (BOOL) drawThemeBackground:(HTHEME)hTheme
		      inRect:(NSRect)rect
		      part:(int)part
		      state:(int)state
{
  HDC hDC;
  RECT winRect;

  if (![self useVisualStyles])
    return NO;

  if (hTheme == NULL)
    return NO;
  
  if ([self isTheme:hTheme partDefined:part] == NO)
    return NO;

  winRect = WIN32RectFromFrame(rect);
  hDC = GetCurrentHDC();

  if (FAILED(DrawThemeBackground(hTheme, hDC, part, state, &winRect, NULL)))
  {
    ReleaseCurrentHDC(hDC);
    return NO;
  }
  else
  {
    ReleaseCurrentHDC(hDC);
    return YES;
  }
}

- (NSSize) defaultSizeForTheme:(HTHEME)hTheme part:(int)part state:(int)state
{
  RECT rect;
  if (FAILED(GetThemeRect(hTheme, part, state, TMT_DEFAULTPANESIZE, &rect)))
    [NSException raise:WIN32ThemeException
                format:@"Unknown error calling GetThemeRect in -[WIN32Theme defaultSizeForTheme:part:state] (hTheme=%p,part=%d,state=%d)",
                hTheme, part, state];
  return NSMakeSize(rect.right - rect.left, rect.bottom - rect.top);
}

- (int) intForTheme:(HTHEME)hTheme part:(int)part state:(int)state 
    property:(int)property
{
  int num;
  if (FAILED(GetThemeInt(hTheme, part, state, property, &num)))
    [NSException raise:WIN32ThemeException
                format:@"Unknown error calling GetThemeInt in -[WIN32Theme intForTheme:part:state:property:] (hTheme=%p,part=%d,state=%d,property=%d)",
                hTheme, part, state, property];
  return num;
}

- (int) metricForTheme:(HTHEME)hTheme part:(int)part state:(int)state 
    property:(int)property
{
  int num;
  HRESULT res;
  
  res = GetThemeMetric(hTheme, NULL, part, state, property, &num);
  if (FAILED(res))
    [NSException raise:WIN32ThemeException
                format:@"Unknown error calling GetThemeMetric in -[WIN32Theme metricForTheme:part:state:property:] (hTheme=%@,part=%d,state=%d,property=%d)",
                [self themeClassName:hTheme],
                part, state, property];
  return num;
}

- (NSRect) backgroundContentRectForTheme:(HTHEME)hTheme
                                    part:(int)part
                                   state:(int)state
                                   frame:(NSRect)frame 
{
  RECT frameRect, rcRect;
  HDC hDC;
  HRESULT result;

  hDC = GetCurrentHDC();
  frameRect = WIN32RectFromFrame(frame);
  result = GetThemeBackgroundContentRect(hTheme, hDC, part, state, &frameRect, &rcRect);

  ReleaseCurrentHDC(hDC);

  if (FAILED(result))
    return NSZeroRect;
  else
    return GSFrameFromWIN32Rect(rcRect);
}

- (WIN32ThemeMargins) marginsForTheme:(HTHEME)hTheme
                                 part:(int)part
                                state:(int)state
{
  WIN32ThemeMargins themeMargins;
  MARGINS win32Margins;
  HRESULT result;


  result = GetThemeMargins(hTheme, NULL, part, state, TMT_CONTENTMARGINS, NULL, &win32Margins);

  if (FAILED(result))
    [NSException raise:WIN32ThemeException
                format:@"Unknown error calling GetThemeMargins in -[WIN32Theme marginsForTheme:part:state:] (hTheme=%p,part=%d,state=%d)",
                hTheme, part, state];


  themeMargins.leftWidth = win32Margins.cxLeftWidth;
  themeMargins.rightWidth = win32Margins.cxRightWidth;
  themeMargins.topHeight = win32Margins.cyTopHeight;
  themeMargins.bottomHeight = win32Margins.cyBottomHeight;
  return themeMargins;
}

- (NSSize) sizeForTheme:(HTHEME)hTheme
                   part:(int)part
                  state:(int)state
                  type:(WIN32ThemeSizeType)sizeType 
{
  int type;
  SIZE size;

  switch (sizeType)
  {
    case WIN32ThemeSizeMinimum: type= TS_MIN; break;
    case WIN32ThemeSizeBestFit: type = TS_TRUE; break;
    case WIN32ThemeSizeDraw: type = TS_DRAW; break;
  }

  if (FAILED(GetThemePartSize(hTheme, NULL, part, state, NULL, type, &size)))
  {
    [NSException raise:WIN32ThemeException
                format:@"Error calling GetThemePartSize in -[WIN32Theme sizeForTheme:part:state:type:] (hTheme=%p,part=%d,state=%d,type=%d",
                hTheme, part, state, sizeType];
    return NSZeroSize;
  }
  else
    return NSMakeSize(size.cx, size.cy);
}
@end

int WindowHeight(HWND window)
{
  RECT rect;

  if (!window)
    {
      NSLog(@"No window for coordinate transformation.");
      return 0;
    }

  if (!GetClientRect(window, &rect))
    {
      NSLog(@"No window rectangle for coordinate transformation.");
      return 0;
    }

  return rect.bottom - rect.top;
}

POINT GSWindowPointToMS( NSPoint p)
{
  POINT p1;
  void *winHandle;
  int x, y;

  [GSCurrentContext() GSCurrentDevice:&winHandle :&x :&y];

  p1.x = p.x - x;
  p1.y = y - p.y;
  
  return p1;
}

RECT GSWindowRectToMS(NSRect r)
{
  void *winHandle;
  int x, y;
  [GSCurrentContext() GSCurrentDevice:&winHandle :&x :&y];

  RECT r1;

  r1.left = r.origin.x - x;
  r1.bottom = y - r.origin.y;
  r1.right = r1.left + r.size.width;
  r1.top = r1.bottom - r.size.height;

  return r1;
}

POINT GSViewPointToWin(NSPoint p)
{
  NSAffineTransform* ctm = [GSCurrentContext() GSCurrentCTM];

  p = [ctm transformPoint: p];
  return GSWindowPointToMS(p);
}

RECT GSViewRectToWin(NSRect r)
{
  NSAffineTransform* ctm = [GSCurrentContext() GSCurrentCTM];

  r = (NSRect)[ctm rectInMatrixSpace: r];
  return GSWindowRectToMS(r);
}

RECT WIN32RectFromFrame(NSRect r)
{
  return GSViewRectToWin(r); 
}


NSRect WIN32RectToGS(RECT rect)
{
  void* winHandle;
  int x, y;
  NSRect output;

  [GSCurrentContext() GSCurrentDevice:&winHandle :&x :&y];

  output.origin.x = rect.left + x;
  /* FIXME: This is most likely incorrect and will stuff up on some systems */
  //output.origin.y = WindowHeight((HWND)winHandle) -  rect.bottom - y;
  output.origin.y = y - rect.bottom;
  output.size.width = rect.right - rect.left;
  output.size.height = rect.bottom - rect.top;
  return output;
}

NSRect WIN32RectToGSViewFrame(RECT rect)
{
  NSAffineTransform *ctm;
  NSRect frame;

  ctm = [[GSCurrentContext() GSCurrentCTM] copy];
  [ctm invert];

  frame = WIN32RectToGS(rect);

  frame = [ctm rectInMatrixSpace:frame];
  [ctm release];
  return frame;
}

NSRect GSFrameFromWIN32Rect(RECT rect)
{
  return WIN32RectToGSViewFrame(rect);
}

HDC GetCurrentHDC()
{
  HDC hdc;
  id gstate;

  gstate = [(id)GSCurrentContext() currentGState]; 
  hdc = (HDC)[gstate getHDC];
  return hdc;
}

void ReleaseCurrentHDC(HDC hdc)
{
  [(id)[(id)GSCurrentContext() currentGState] releaseHDC:hdc];
}

NSString* WIN32StringFromRect(RECT rect)
{
  return [NSString stringWithFormat:@"{left=%d, right=%d, top=%d, bottom=%d}",
    rect.left, rect.right, rect.top, rect.bottom];
}


